@@ -124,6 +124,8 @@ struct tasklet_struct {
 #define tasklet_unlock_wait(t)	while (test_bit(0, &(t)->sync)) { }
 #define tasklet_kill(t)		tasklet_unlock_wait(t) /* FIXME: world is not perfect... */
 
+#define del_timer_sync(t) del_timer(t) /* FIXME: not quite correct on SMP */
+
 #define rwlock_init(x) do { *(x) = RW_LOCK_UNLOCKED; } while(0)
 
 #ifndef __init
@@ -320,15 +322,7 @@ void snd_pci_compat_release_regions(struct pci_dev *pdev);
 void snd_pci_compat_save_state(struct pci_dev *pdev, u32 *buf);
 void snd_pci_compat_restore_state(struct pci_dev *pdev, u32 *buf);
 
-static inline int pci_module_init(struct pci_driver *drv)
-{
-	int res = snd_pci_compat_register_driver(drv);
-	if (res < 0)
-		return res;
-	if (res == 0)
-		return -ENODEV;
-	return 0;
-}
+#define pci_module_init	snd_pci_compat_register_driver
 
 #endif /* CONFIG_PCI */
 
@@ -479,6 +473,14 @@ extern inline int pm_send(struct pm_dev *dev, pm_request_t rqst, void *data)
 
 #include <linux/usb.h>
 
+struct usb_ctrlrequest {
+	__u8 bRequestType;
+	__u8 bRequest;
+	__u16 wValue;
+	__u16 wIndex;
+	__u16 wLength;
+} __attribute__ ((packed));
+
 struct snd_compat_usb_device_id {
 	__u16		match_flags;
 	__u16		idVendor;
