@@ -97,11 +97,13 @@ static char *hiddendir = "..";
 static char *kernel_deps[] = {
 	/* buses */
 	"ISA",
+	"ISA_DMA_API",
 	"ISAPNP",
 	"EISA",
 	"PCI",
 	"SBUS",
 	"I2C",
+	"INPUT",
 	"L3",
 	"USB",
 	"PCMCIA",
@@ -112,53 +114,66 @@ static char *kernel_deps[] = {
 	"SPARC64",
 	"PPC",
 	"PPC64",
+	"PPC_PMAC",
 	"X86_64",
 	"X86",
+	"MIPS",
 	"MIPS64",
 	"IA32_EMULATION",
 	/* architecture specific */
 	"ARCH_SA1100",
+	"ARCH_PXA",
 	"X86_PC9800",
 	/* other drivers */
 	"RTC",
 	"GAMEPORT",
 	"VIDEO_DEV",
 	"FW_LOADER",
-	"ASIHPK", /* external tree */
 	/* some flags */
 	"EXPERIMENTAL",
 	NULL
 };
 
 /* % -> always true */
+/* # -> define */
 static char *no_cards[] = {
 	"%SOUND",
 	"SOUND_PRIME",
 	"%SND",
-	"SND_TIMER",
-	"SND_HWDEP",
-	"SND_RAWMIDI",
-	"SND_PCM",
+	"#SND_TIMER",
+	"#SND_HWDEP",
+	"#SND_RAWMIDI",
+	"#SND_PCM",
 	"SND_SEQUENCER",
 	"SND_MIXER_OSS",
 	"SND_PCM_OSS",
 	"SND_SEQUENCER_OSS",
 	"SND_OSSEMUL",
 	"SND_RTCTIMER",
+	"SND_HPET",
 	"SND_DEBUG",
 	"SND_DEBUG_MEMORY",
 	"SND_DEBUG_DETECT",
 	"SND_VERBOSE_PRINTK",
 	"SND_BIT32_EMUL",
-	"SND_OPL3_LIB",
-	"SND_OPL4_LIB",
-	"SND_VX_LIB",
-	"SND_MPU401_UART",
-	"SND_GUS_SYNTH",
-	"SND_AC97_CODEC",
+	"#SND_OPL3_LIB",
+	"#SND_OPL4_LIB",
+	"#SND_VX_LIB",
+	"#SND_MPU401_UART",
+	"#SND_GUS_SYNTH",
+	"#SND_AC97_BUS",
+	"#SND_AC97_CODEC",
+	"#SND_PXA2XX_PCM",
+	"#SND_GENERIC_PM",
 	NULL
 };
 
+static const char *kernel26_opts[] = {
+	"SND_AC97_BUS",
+	"SND_PXA2XX_PCM",
+	NULL,
+};
+
 #define READ_STATE_NONE		0
 #define READ_STATE_CONFIG	1
 #define READ_STATE_MENU		2
@@ -287,9 +302,9 @@ static int read_file_1(const char *filename, struct cond **template)
 		}
 		switch (state) {
 		case READ_STATE_CONFIG:
-			if (!strncmp(buffer, "\ttristate ", 10))
+			if (!strncmp(buffer, "\ttristate", 9))
 				dep->is_bool = 0;
-			else if (!strncmp(buffer, "\tbool ", 6))
+			else if (!strncmp(buffer, "\tbool", 5))
 				dep->is_bool = 1;
 			else if (!strncmp(buffer, "\tdepends on ", 12))
 				add_dep(dep, buffer + 12, *template);
@@ -767,7 +782,7 @@ static int is_toplevel(struct dep *dep)
 	}
 	for (idx = 0; no_cards[idx]; idx++) {
 		str = no_cards[idx];
-		if (*str == '%')
+		if (*str == '%' || *str == '#')
 			str++;
 		if (!strcmp(str, dep->name))
 			return 0;
@@ -775,6 +790,33 @@ static int is_toplevel(struct dep *dep)
 	return 1;
 }
 
+// whether to output AC_DEFINE() for this variable
+static int output_ac_define(struct dep *dep)
+{
+	int idx;
+	char *str;
+	
+	if (dep == NULL)
+		return 0;
+	for (idx = 0; kernel_deps[idx]; idx++) {
+		if (!strcmp(kernel_deps[idx], dep->name))
+			return 0;
+	}
+	for (idx = 0; no_cards[idx]; idx++) {
+		int def = 0;
+		str = no_cards[idx];
+		if (*str == '%')
+			str++;
+		if (*str == '#') {
+			def = 1;
+			str++;
+		}
+		if (!strcmp(str, dep->name))
+			return def;
+	}
+	return 1;
+}
+
 // is CONFIG_ variable is always true
 static int is_always_true(struct dep *dep)
 {
@@ -847,6 +889,17 @@ static void sel_remove_hitflags(void)
 	}
 }
 
+// is 2.6-only config option
+static int is_kernel26(const char *name)
+{
+	const char **p;
+	for (p = kernel26_opts; *p; p++)
+		if (! strcmp(name, *p))
+			return 1;
+	return 0;
+}
+
+
 static void sel_print_acinclude(struct sel *sel)
 {
 	struct dep * dep;
@@ -864,7 +917,10 @@ static void sel_print_acinclude(struct sel *sel)
 		sel_print_acinclude(nsel);
 		if (!nsel->dep->hitflag) {
 			nsel->dep->hitflag = 1;
-			printf("      CONFIG_%s=\"%c\"\n", nsel->name,
+			printf("      ");
+			if (is_kernel26(nsel->name))
+				printf("test $kpatchlevel = 6 && ");
+			printf("CONFIG_%s=\"%c\"\n", nsel->name,
 				(nsel->dep && nsel->dep->is_bool) ? 'y' : 'm');
 		}
 	}
@@ -943,11 +999,14 @@ static void output_acinclude(void)
 		for (sel = tempdep->sel; sel; sel = sel->next) {
 			if (is_always_true(sel->dep))
 				continue;
-			printf("      CONFIG_%s=\"%c\"\n", sel->name,
+			printf("      ");
+			if (is_kernel26(sel->name))
+				printf("test $kpatchlevel = 6 && ");
+			printf("CONFIG_%s=\"%c\"\n", sel->name,
 			       (sel->dep && sel->dep->is_bool) ? 'y' : 'm');
 		}
 		text = convert_to_config_uppercase("CONFIG_", tempdep->name);
-		printf("      %s=\"m\"\n", text);
+		printf("      %s=\"%c\"\n", text, tempdep->is_bool ? 'y' : 'm');
 		free(text);
 		printf("      probed=1\n");
 		if (put_if)
@@ -1011,7 +1070,10 @@ static void output_acinclude(void)
 		for (sel = tempdep->sel; sel; sel = sel->next) {
 			if (is_always_true(sel->dep))
 				continue;
-			printf("      CONFIG_%s=\"%c\"\n", sel->name,
+			printf("      ");
+			if (is_kernel26(sel->name))
+				printf("test $kpatchlevel = 6 && ");
+			printf("CONFIG_%s=\"%c\"\n", sel->name,
 			       (sel->dep && sel->dep->is_bool) ? 'y' : 'm');
 		}
 		text = convert_to_config_uppercase("CONFIG_", tempdep->name);
@@ -1034,7 +1096,7 @@ static void output_acinclude(void)
 	printf("AC_MSG_RESULT($SELECTED_CARDS)\n\n");
 	printf("CONFIG_SND=\"m\"\n");
 	for (tempdep = all_deps; tempdep; tempdep = tempdep->next) {
-		if (!is_toplevel(tempdep))
+		if (!output_ac_define(tempdep))
 			continue;
 		text = convert_to_config_uppercase("CONFIG_", tempdep->name);
 		printf("if test -n \"$%s\"; then\n", text);
